import numpy
import pytest
import os
from shutil import rmtree
from numpy.testing import assert_allclose, assert_array_equal
from fgivenx.samples import compute_samples, samples_from_getdist_chains

try:
    import getdist

    def test_samples_from_getdist_chains():

        # Set up getdist chains
        file_root = './.chains/test'
        labels = [r'\alpha', r'\beta', r'\gamma']
        names = ['a', 'b', 'g']
        nsamples = 1000
        params = ['a', 'g']
        i = [names.index(p) for p in params]
        numpy.random.seed(0)
        samples_ = numpy.random.rand(nsamples, len(names))
        weights_ = numpy.random.rand(nsamples)
        samples = getdist.mcsamples.MCSamples(samples=samples_, labels=labels,
                                              names=names, weights=weights_)
        samples.saveAsText(file_root, make_dirs=True)

        samples, weights = samples_from_getdist_chains(params, file_root)
        assert_allclose(samples, samples_[:, i])
        assert_allclose(weights, weights_)

        samples, weights, latex = samples_from_getdist_chains(params,
                                                              file_root,
                                                              latex=True)
        assert_allclose(weights, weights_)
        assert_array_equal(latex, numpy.array(labels)[i])

        settings = {'ignore_rows': 0.5}
        samples1, weights = samples_from_getdist_chains(params, file_root,
                                                        settings=settings)
        assert len(samples1) < len(samples)

        rmtree('./.chains')

except ImportError:
    pass


def test_compute_samples():

    with pytest.raises(TypeError):
        compute_samples(None, None, None, wrong_argument=None)

    cache = '.test_cache/test'
    numpy.random.seed(0)
    nsamp = 5000
    a, b, e, f = 0, 1, 0, 1
    m = numpy.random.normal(a, b, nsamp)
    c = numpy.random.normal(e, f, nsamp)
    samples = numpy.array([list(_) for _ in zip(m, c)])
    nx = 100
    x = numpy.linspace(-1, 1, nx)
    fsamps_ = numpy.outer(x, m) + c

    def f(x, theta):
        m, c = theta
        return m*x+c

    assert(not os.path.isfile(cache + '_fsamples.pkl'))
    fsamps = compute_samples([f], x, [samples], cache=cache)
    assert(os.path.isfile(cache + '_fsamples.pkl'))

    assert_allclose(fsamps_, fsamps,)

    fsamps = compute_samples([f], x, [samples], cache=cache)
    assert_allclose(fsamps_, fsamps,)

    rmtree('.test_cache')



import numpy
from fgivenx.parallel import parallel_apply
from fgivenx.io import CacheException, Cache


def compute_samples(f, x, samples, **kwargs):
    r""" Apply f(x,theta) to x array and theta in samples.

    Parameters
    ----------
    f: function
        list of functions :math:`f(x;\theta)`  with dependent variable
        :math:`x`, parameterised by :math:`\theta`.

    x: 1D array-like
        x values to evaluate :math:`f(x;\theta)` at.

    samples: 2D array-like
        list of theta samples to evaluate :math:`f(x;\theta)` at.
        `shape = (nfunc, nsamples, npars)`

    parallel, tqdm_kwargs: optional
        see docstring for :func:`fgivenx.parallel.parallel_apply`

    cache: str, optional
        File root for saving previous calculations for re-use
        default None

    Returns
    -------
    2D numpy.array:
        samples at each x. `shape=(len(x),len(samples),)`
    """

    parallel = kwargs.pop('parallel', False)
    cache = kwargs.pop('cache', '')
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})
    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    if cache:
        cache = Cache(cache + '_fsamples')
        try:
            return cache.check(x, samples)
        except CacheException as e:
            print(e)

    fsamples = []
    for fi, s in zip(f, samples):
        if len(s) > 0:
            fsamps = parallel_apply(fi, s, precurry=(x,), parallel=parallel,
                                    tqdm_kwargs=tqdm_kwargs)
            fsamps = numpy.array(fsamps).transpose().copy()
            fsamples.append(fsamps)
    fsamples = numpy.concatenate(fsamples, axis=1)

    if cache:
        cache.save(x, samples, fsamples)

    return fsamples


def samples_from_getdist_chains(params, file_root, latex=False, **kwargs):
    """ Extract samples and weights from getdist chains.

    Parameters
    ----------
    params: list(str)
        Names of parameters to be supplied to second argument of f(x|theta).

    file_root: str, optional
        Root name for getdist chains files. This variable automatically
        defines:
        - chains_file = file_root.txt
        - paramnames_file = file_root.paramnames
        but can be overidden by chains_file or paramnames_file.

    latex: bool, optional
        Also return an array of latex strings for those paramnames.

    Any additional keyword arguments are forwarded onto getdist, e.g:

        samples_from_getdist_chains(params, file_root,
                                    settings={'ignore_rows':0.5})

    Returns
    -------
    samples: numpy.array
        2D Array of samples. `shape=(len(samples), len(params))`

    weights: numpy.array
        Array of weights. `shape = (len(params),)`

    latex: list(str), optional
        list of latex strings for each parameter
        (if latex is provided as an argument)
    """

    import getdist
    samples = getdist.loadMCSamples(file_root, **kwargs)
    weights = samples.weights

    indices = [samples.index[p] for p in params]
    samps = samples.samples[:, indices]
    if latex:
        latex = [samples.parLabel(p) for p in params]
        return samps, weights, latex
    else:
        return samps, weights



import numpy
import pytest
import os
from shutil import rmtree
from numpy.testing import assert_allclose
import scipy.stats
import scipy.integrate
import scipy.special
from fgivenx.mass import PMF, compute_pmf


def gaussian_pmf(y, mu=0, sigma=1):
    return scipy.special.erfc(numpy.abs(y-mu)/numpy.sqrt(2)/sigma)


def test_gaussian():
    numpy.random.seed(0)
    nsamp = 5000
    samples = numpy.random.randn(nsamp)
    y = numpy.random.uniform(-3, 3, 10)
    m = PMF(samples, y)
    m_ = gaussian_pmf(y)
    assert_allclose(m, m_, rtol=3e-1)


def test_PMF():
    # Compute samples
    numpy.random.seed(0)
    nsamp = 100
    samples = numpy.concatenate((-5+numpy.random.randn(nsamp//2),
                                 5+numpy.random.randn(nsamp//2)))

    # Compute PMF
    y = numpy.random.uniform(-10, 10, 10)
    m = PMF(samples, y)

    # Compute PMF via monte carlo
    N = 100000
    kernel = scipy.stats.gaussian_kde(samples)
    s = kernel.resample(N)[0]
    m_ = [sum(kernel(s) <= kernel(y_i))/float(N) for y_i in y]
    assert_allclose(m, m_, atol=3*N**-0.5)

    # Compute PMF via quadrature
    m_ = [scipy.integrate.quad(lambda x: kernel(x)*(kernel(x) <= kernel(y_i)),
                               -numpy.inf, numpy.inf, limit=500)[0]
          for y_i in y]
    assert_allclose(m, m_, atol=1e-4)

    assert_allclose([0, 0], PMF(samples, [-1e3, 1e3]))

    samples = [0, 0]
    m = PMF(samples, y)
    assert_allclose(m, numpy.zeros_like(y))


def test_compute_pmf():

    with pytest.raises(TypeError):
        compute_pmf(None, None, wrong_argument=None)

    cache = '.test_cache/test'
    numpy.random.seed(0)
    nsamp = 5000
    a, b, e, f = 0, 1, 0, 1
    m = numpy.random.normal(a, b, nsamp)
    c = numpy.random.normal(e, f, nsamp)
    nx = 100
    x = numpy.linspace(-1, 1, nx)
    fsamps = (numpy.outer(x, m) + c)
    ny = 100
    y = numpy.linspace(-3, 3, ny)

    assert(not os.path.isfile(cache + '_masses.pkl'))
    m = compute_pmf(fsamps, y, cache=cache)
    assert(os.path.isfile(cache + '_masses.pkl'))

    m_ = [gaussian_pmf(y, a*xi+e, numpy.sqrt(b**2*xi**2+f**2)) for xi in x]
    assert_allclose(m.transpose(), m_, atol=3e-1)

    m = compute_pmf(fsamps, y, cache=cache)
    assert_allclose(m.transpose(), m_, atol=3e-1)

    rmtree('.test_cache')



import numpy
import pytest
from numpy.testing import assert_allclose, assert_almost_equal
from fgivenx._utils import _check_args, _normalise_weights, \
                           _equally_weight_samples


def test__check_args():
    numpy.random.seed(0)
    nfuncs = 3
    logZ = numpy.random.rand(nfuncs)
    f = [lambda x: x**i for i in range(nfuncs)]
    nx = 100
    x = numpy.linspace(0, 1, nx)
    nsamps = 200
    nparams = 5
    samples = numpy.random.rand(nfuncs, nsamps, nparams)
    weights = numpy.random.rand(nfuncs, nsamps)

    # check these valid versions pass
    _check_args(logZ, f, x, samples, weights)
    _check_args(None, f[0], x, samples[0], weights[0])

    with pytest.raises(ValueError):
        _check_args(numpy.ones((2, 2)), f, x, samples, weights)

    with pytest.raises(ValueError):
        _check_args(logZ, f, numpy.ones((2, 2)), samples, weights)

    with pytest.raises(ValueError):
        _check_args(logZ, f, numpy.ones((2, 2)), samples, weights)

    with pytest.raises(ValueError):
        _check_args(logZ, f[1:], x, samples, weights)

    with pytest.raises(ValueError):
        _check_args(logZ, numpy.ones_like(f), x, samples, weights)

    with pytest.raises(ValueError):
        _check_args(logZ, f, x, samples[1:], weights)

    with pytest.raises(ValueError):
        _check_args(logZ, f, x, numpy.random.rand(nfuncs, nparams), weights)

    with pytest.raises(ValueError):
        _check_args(logZ, f, x, samples, weights[1:])

    with pytest.raises(ValueError):
        _check_args(logZ, f, x, samples, numpy.random.rand(nfuncs))

    with pytest.raises(ValueError):
        _check_args(logZ, f, x, samples, numpy.random.rand(nfuncs, nsamps+1))


def assert_in_ratio(a, b):
    assert_almost_equal(numpy.array(a)*max(b), max(a)*numpy.array(b))


def test__normalise_weights():
    numpy.random.seed(0)
    nfuncs = 5
    nsamps = 5000
    ntrim = 500
    logZ = numpy.random.rand(nfuncs)
    weights = numpy.random.rand(nfuncs, nsamps)

    logZ, weights = _normalise_weights(logZ, weights)

    assert_almost_equal(numpy.max(weights), 1)
    assert_almost_equal(numpy.max(logZ), 0)
    assert_in_ratio([sum(w) for w in weights], numpy.exp(logZ))

    logZ, weights = _normalise_weights(logZ, weights, ntrim)

    assert_in_ratio([sum(w) for w in weights], numpy.exp(logZ))
    assert_almost_equal(numpy.sum(weights), ntrim)


def test__equally_weight_samples():
    numpy.random.seed(0)
    nsamps = 5000
    nparams = 5
    samples = numpy.random.rand(nsamps, nparams)
    weights = numpy.random.rand(nsamps)

    with pytest.raises(ValueError):
        _equally_weight_samples(samples, weights[1:])

    with pytest.raises(ValueError):
        _equally_weight_samples(samples, weights+1)

    with pytest.raises(ValueError):
        _equally_weight_samples(samples, weights-1)

    samples_1 = _equally_weight_samples(samples, weights)
    rand_1 = numpy.random.rand()
    samples_2 = _equally_weight_samples(samples, weights)
    rand_2 = numpy.random.rand()

    assert_allclose(samples_1, samples_2)
    assert(rand_1 != rand_2)
    assert_almost_equal(len(samples_1)/sum(weights), 1, 1)



import scipy
import scipy.ndimage
import numpy
import matplotlib.pyplot


def plot(x, y, z, ax=None, **kwargs):
    r"""
    Plot iso-probability mass function, converted to sigmas.

    Parameters
    ----------
    x, y, z : numpy arrays
        Same as arguments to :func:`matplotlib.pyplot.contour`

    ax: axes object, optional
        :class:`matplotlib.axes._subplots.AxesSubplot` to plot the contours
        onto. If unsupplied, then :func:`matplotlib.pyplot.gca()` is used to
        get the last axis used, or create a new one.

    colors: color scheme, optional
        :class:`matplotlib.colors.LinearSegmentedColormap`
        Color scheme to plot with. Recommend plotting in reverse
        (Default: :class:`matplotlib.pyplot.cm.Reds_r`)

    alpha: float, optional
        Transparency of filled contours. Given as alpha blending
        value between 0 (transparent) and 1 (opague).

    smooth: float, optional
        Percentage by which to smooth the contours.
        (Default: no smoothing)

    contour_line_levels: List[float], optional
        Contour lines to be plotted.  (Default: [1,2])

    linewidths: float, optional
        Thickness of contour lines.  (Default: 0.3)

    contour_color_levels: List[float], optional
        Contour color levels. (Default:
        `numpy.arange(0, contour_line_levels[-1] + fineness / 2, fineness)`)

    fineness: float, optional
        Spacing of contour color levels.  (Default: 0.5)

    lines: bool, optional
        (Default: True)

    rasterize_contours: bool, optional
        Rasterize the contours while keeping the lines, text etc in vector
        format. Useful for reducing file size bloat and making printing
        easier when you have dense contours.
        (Default: False)

    Returns
    -------
    cbar: color bar
        :class:`matplotlib.contour.QuadContourSet`
        Colors to create a global colour bar
    """
    if ax is None:
        ax = matplotlib.pyplot.gca()
    # Get inputs
    colors = kwargs.pop('colors', matplotlib.pyplot.cm.Reds_r)
    smooth = kwargs.pop('smooth', False)

    linewidths = kwargs.pop('linewidths', 0.3)
    contour_line_levels = kwargs.pop('contour_line_levels', [1, 2, 3])

    fineness = kwargs.pop('fineness', 0.5)
    default_color_levels = numpy.arange(0,
                                        contour_line_levels[-1] + fineness / 2,
                                        fineness)
    contour_color_levels = kwargs.pop('contour_color_levels',
                                      default_color_levels)

    rasterize_contours = kwargs.pop('rasterize_contours', False)

    alpha = kwargs.pop('alpha', 1)

    lines = kwargs.pop('lines', True)

    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    # Convert to sigmas
    z = numpy.sqrt(2) * scipy.special.erfinv(1 - z)

    # Gaussian filter if desired the sigmas by a factor of smooth%
    if smooth:
        sigma = smooth*numpy.array(z.shape)/100.0
        z = scipy.ndimage.gaussian_filter(z, sigma=sigma, order=0)

    # Plot the filled contours onto the axis ax
    cbar = ax.contourf(x, y, z, cmap=colors, levels=contour_color_levels,
                       alpha=alpha)

    # Rasterize contours (the rest of the figure stays in vector format)
    if rasterize_contours:
        for c in cbar.collections:
            c.set_rasterized(True)

    # Remove those annoying white lines
    for c in cbar.collections:
        c.set_edgecolor("face")

    # Plot some sigma-based contour lines
    if lines:
        ax.contour(x, y, z, colors='k', linewidths=linewidths,
                   levels=contour_line_levels)

    # Return the contours for use as a colourbar later
    return cbar


def plot_lines(x, fsamps, ax=None, downsample=100, **kwargs):
    """
    Plot function samples as a set of line plots.

    Parameters
    ----------
    x: 1D array-like
        x values to plot

    fsamps: 2D array-like
        set of functions to plot at each x. As returned by
        :func:`fgivenx.compute_samples`

    ax: axes object
        :class:`matplotlib.pyplot.ax` to plot on.

    downsample: int, optional
        Reduce the number of samples to a viewable quantity. (Default 100)

    any other keywords are passed to :meth:`matplotlib.pyplot.ax.plot`
    """
    if ax is None:
        ax = matplotlib.pyplot.gca()
    if downsample < len(fsamps.T):
        indices = numpy.random.choice(len(fsamps.T), downsample, replace=False)
    else:
        indices = numpy.arange(len(fsamps.T))
    color = kwargs.pop('color', 'k')
    alpha = kwargs.pop('alpha', 0.1)
    for y in fsamps.T[indices]:
        ax.plot(x, y, color=color, alpha=alpha, **kwargs)



import numpy
import pytest
import os
import fgivenx.io
from shutil import rmtree
from numpy.testing import assert_allclose


def test_Cache():
    """ Fully test the cache """

    # Stuff for testing the functionality
    numpy.random.seed(0)
    dirname = '.test_cache'
    root = os.path.join(dirname, 'test_root')
    data0 = numpy.random.rand(5, 4, 3)
    data1 = numpy.random.rand(4, 3)
    data2 = data0 * data1

    # Creates a directory
    assert(not os.path.isdir(dirname))
    cache = fgivenx.io.Cache(root)
    assert(os.path.isdir(dirname))

    # Raises exception if there's no cache
    with pytest.raises(fgivenx.io.CacheMissing):
        cache.load()

    # Check saving
    cache.save(data0, data1, data2)
    assert(os.path.exists(root + '.pkl'))

    # check loading
    data0_, data1_, data2_ = cache.load()
    assert_allclose(data0_, data0)
    assert_allclose(data1_, data1)
    assert_allclose(data2_, data2)

    # check check
    with pytest.raises(ValueError):
        cache.check(data0)

    with pytest.raises(fgivenx.io.CacheChanged):
        cache.check(data1, data0)

    with pytest.raises(fgivenx.io.CacheChanged):
        cache.check(numpy.random.rand(5, 4, 3), numpy.random.rand(4, 3))

    assert_allclose(data2, cache.check(data0, data1))

    cache.save([data0, data1], data2)

    assert_allclose(data2, cache.check([data0_, data1_]))

    with pytest.raises(fgivenx.io.CacheChanged):
        cache.check([data1])

    with pytest.raises(fgivenx.io.CacheChanged):
        cache.check([data1, data0])

    with pytest.raises(fgivenx.io.CacheChanged):
        cache.check([data0, 1+data1])

    # Remove the testing cache
    rmtree(dirname)



import pytest
import numpy
import matplotlib.pyplot as plt
from fgivenx import plot_contours, plot_lines, plot_dkl
from fgivenx.drivers import compute_samples, compute_pmf, compute_dkl


def test_full():
    # Model definitions
    # =================
    # Define a simple straight line function, parameters theta=(m,c)
    def f(x, theta):
        m, c = theta
        return m * x + c

    numpy.random.seed(1)

    # Posterior samples
    nsamples = 1000
    ms = numpy.random.normal(loc=-5, scale=1, size=nsamples)
    cs = numpy.random.normal(loc=2, scale=1, size=nsamples)
    samples = numpy.array([(m, c) for m, c in zip(ms, cs)]).copy()

    # Prior samples
    ms = numpy.random.normal(loc=0, scale=5, size=nsamples)
    cs = numpy.random.normal(loc=0, scale=5, size=nsamples)
    prior_samples = numpy.array([(m, c) for m, c in zip(ms, cs)]).copy()

    # Computation
    # ===========
    # Examine the function over a range of x's
    xmin, xmax = -2, 2
    nx = 100
    x = numpy.linspace(xmin, xmax, nx)

    # Set the cache
    cache = 'cache/test'
    prior_cache = cache + '_prior'

    # Compute function samples
    with pytest.raises(TypeError):
        compute_samples(f, x, samples, unknown_arg=None)
    compute_samples(f, x, samples, cache=cache)
    compute_samples(f, x, prior_samples, cache=prior_cache)

    # Compute dkls
    with pytest.raises(TypeError):
        compute_dkl(f, x, samples, prior_samples, unknown_arg=None)
    compute_dkl(f, x, samples, prior_samples,
                cache=cache, parallel=True)

    # Compute probability mass function.
    with pytest.raises(TypeError):
        compute_pmf(f, x, samples, unknown_arg=None)
    with pytest.raises(ValueError):
        compute_pmf(f, x, samples, y=numpy.random.rand(2, 2))
    y, pmf = compute_pmf(f, x, samples, cache=cache, parallel=True)
    y_prior, pmf_prior = compute_pmf(f, x, prior_samples,
                                     cache=prior_cache, parallel=True)

    plot_dkl(f, x, samples, prior_samples,
             cache=cache, prior_cache=prior_cache)


def test_plotting():
    # Model definitions
    # =================
    # Define a simple straight line function, parameters theta=(m,c)
    def f(x, theta):
        m, c = theta
        return m * x + c

    numpy.random.seed(1)

    # Posterior samples
    nsamples = 1000
    ms = numpy.random.normal(loc=-5, scale=1, size=nsamples)
    cs = numpy.random.normal(loc=2, scale=1, size=nsamples)
    samples = numpy.array([(m, c) for m, c in zip(ms, cs)]).copy()

    # Prior samples
    ms = numpy.random.normal(loc=0, scale=5, size=nsamples)
    cs = numpy.random.normal(loc=0, scale=5, size=nsamples)
    prior_samples = numpy.array([(m, c) for m, c in zip(ms, cs)]).copy()

    # Examine the function over a range of x's
    xmin, xmax = -2, 2
    nx = 100
    x = numpy.linspace(xmin, xmax, nx)

    # Set the cache
    for cache in [None, 'cache/test']:
        if cache is not None:
            prior_cache = cache + '_prior'
        else:
            prior_cache = None

        # Plotting
        # ========
        fig, axes = plt.subplots(2, 2)

        # Sample plot
        # -----------
        ax_samples = axes[0, 0]
        ax_samples.set_ylabel(r'$c$')
        ax_samples.set_xlabel(r'$m$')
        ax_samples.plot(prior_samples.T[0], prior_samples.T[1], 'b.')
        ax_samples.plot(samples.T[0], samples.T[1], 'r.')

        # Line plot
        # ---------
        ax_lines = axes[0, 1]
        ax_lines.set_ylabel(r'$y = m x + c$')
        ax_lines.set_xlabel(r'$x$')
        plot_lines(f, x, prior_samples, ax_lines, color='b', cache=prior_cache)
        plot_lines(f, x, samples, ax_lines, color='r', cache=cache)

        # Predictive posterior plot
        # -------------------------
        ax_fgivenx = axes[1, 1]
        ax_fgivenx.set_ylabel(r'$P(y|x)$')
        ax_fgivenx.set_xlabel(r'$x$')
        plot_contours(f, x, prior_samples, ax_fgivenx,
                      colors=plt.cm.Blues_r, lines=False,
                      cache=prior_cache)
        plot_contours(f, x, samples, ax_fgivenx, cache=cache)

        # DKL plot
        # --------
        ax_dkl = axes[1, 0]
        ax_dkl.set_ylabel(r'$D_\mathrm{KL}$')
        ax_dkl.set_xlabel(r'$x$')
        ax_dkl.set_ylim(bottom=0)
        plot_dkl(f, x, samples, prior_samples, ax_dkl,
                 cache=cache, prior_cache=prior_cache)

        ax_lines.get_shared_x_axes().join(ax_lines, ax_fgivenx, ax_samples)
        fig.set_size_inches(6, 6)



import warnings
try:
    from tqdm import tqdm, tqdm_notebook
except ImportError:
    def tqdm(x, **kwargs): return x

    def tqdm_notebook(x, **kwargs): return x


try:
    PARALLEL = True
    from joblib import Parallel, delayed, cpu_count
except ImportError:
    PARALLEL = False

    class Parallel(object):
        def __init__(self, n_jobs=None): pass

        def __call__(self, x): return list(x)

    def delayed(x): return x

    def cpu_count(): return 1


def parallel_apply(f, array, **kwargs):
    """ Apply a function to an array with openmp parallelisation.

    Equivalent to `[f(x) for x in array]`, but parallelised if required.

    Parameters
    ----------
    f: function
        Univariate function to apply to each element of array

    array: array-like
        Array to apply f to

    parallel: int or bool, optional
        int > 0: number of processes to parallelise over

        int < 0 or bool=True: use OMP_NUM_THREADS to choose parallelisation

        bool=False or int=0: do not parallelise

    tqdm_kwargs: dict, optional
        additional kwargs for tqdm progress bars.

    precurry: tuple, optional
        immutable arguments to pass to f before x,
        i.e. `[f(precurry,x) for x in array]`

    postcurry: tuple, optional
        immutable arguments to pass to f after x
        i.e. `[f(x,postcurry) for x in array]`

    Returns
    -------
    list:
        `[f(precurry,x,postcurry) for x in array]`
        parallelised according to parallel
    """

    precurry = tuple(kwargs.pop('precurry', ()))
    postcurry = tuple(kwargs.pop('postcurry', ()))
    parallel = kwargs.pop('parallel', False)
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})
    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)
    try:
        # If running in a jupyter notebook then use tqdm_notebook.
        progress = tqdm_notebook if get_ipython().has_trait('kernel') else tqdm
    except (NameError, AssertionError):
        # Otherwise use regular tqdm progress bar
        progress = tqdm
    if not parallel:
        return [f(*(precurry + (x,) + postcurry)) for x in
                progress(array, **tqdm_kwargs)]
    elif parallel is True:
        parallel = cpu_count()
    elif isinstance(parallel, int):
        if parallel < 0:
            parallel = cpu_count()
        else:
            parallel = parallel
    else:
        raise ValueError("parallel keyword must be an integer or bool")

    if parallel and not PARALLEL:
        warnings.warn("You need to install the package joblib"
                      "if you want to use parallelisation")

    return Parallel(n_jobs=parallel)(delayed(f)(*(precurry + (x,) + postcurry))
                                     for x in progress(array, **tqdm_kwargs))



import numpy


def _check_args(logZ, f, x, samples, weights):
    """ Sanity-check the arguments for :func:`fgivenx.drivers.compute_samples`.

    Parameters
    ----------
    f, x, samples, weights:
        see arguments for :func:`fgivenx.drivers.compute_samples`
    """
    # convert to arrays
    if logZ is None:
        logZ = [0]
        f = [f]
        samples = [samples]
        weights = [weights]

    # logZ
    logZ = numpy.array(logZ, dtype='double')
    if len(logZ.shape) != 1:
        raise ValueError("logZ should be a 1D array")

    # x
    x = numpy.array(x, dtype='double')
    if len(x.shape) != 1:
        raise ValueError("x should be a 1D array")

    # f
    if len(logZ) != len(f):
        raise ValueError("len(logZ) = %i != len(f) = %i"
                         % (len(logZ), len(f)))
    for func in f:
        if not callable(func):
            raise ValueError("first argument f must be function"
                             "(or list of functions) of two variables")

    # samples
    if len(logZ) != len(samples):
        raise ValueError("len(logZ) = %i != len(samples) = %i"
                         % (len(logZ), len(samples)))
    samples = [numpy.array(s, dtype='double') for s in samples]
    for s in samples:
        if len(s.shape) != 2:
            raise ValueError("each set of samples should be a 2D array")

    # weights
    if len(logZ) != len(weights):
        raise ValueError("len(logZ) = %i != len(weights) = %i"
                         % (len(logZ), len(weights)))
    weights = [numpy.array(w, dtype='double') if w is not None
               else numpy.ones(len(s), dtype='double')
               for w, s in zip(weights, samples)]

    for w, s in zip(weights, samples):
        if len(w.shape) != 1:
            raise ValueError("each set of weights should be a 1D array")
        if len(w) != len(s):
            raise ValueError("len(w) = %i != len(s) = %i" % (len(s), len(w)))

    return logZ, f, x, samples, weights


def _normalise_weights(logZ, weights, ntrim=None):
    """ Correctly normalise the weights for trimming

    This takes a list of log-evidences, and re-normalises the weights so that
    the largest weight across all samples is 1, and the total weight in each
    set of samples is proportional to the evidence.

    Parameters
    ----------
    logZ: array-like
        log-evidences to weight each set of weights by

    weights: array-like of numpy.array
        list of not necessarily equal length list of weights

    Returns
    -------
    logZ: numpy.array
        evidences, renormalised so that max(logZ) = 0

    weights: list of 1D numpy.array
        normalised weights
    """
    logZ -= logZ.max()

    Zs = numpy.exp(logZ)

    weights = [w/w.sum()*Z for w, Z in zip(weights, Zs)]

    wmax = max([w.max() for w in weights])

    weights = [w/wmax for w in weights]

    ntot = sum([w.sum() for w in weights])

    if ntrim is not None and ntrim < ntot:
        weights = [w*ntrim/ntot for w in weights]

    return logZ, weights


def _equally_weight_samples(samples, weights):
    """ Convert samples to be equally weighted.

    Samples are trimmed by discarding samples in accordance with a probability
    determined by the corresponding weight.

    This function has assumed you have normalised the weights properly.
    If in doubt, convert weights via: `weights /= weights.max()`

    Parameters
    ----------
    samples: array-like
        Samples to trim.

    weights: array-like
        Weights to trim by.

    Returns
    -------
    1D numpy.array:
        Equally weighted sample array. `shape=(len(samples))`
    """
    if len(weights) != len(samples):
        raise ValueError("len(weights) = %i != len(samples) = %i" %
                         (len(weights), len(samples)))

    if numpy.logical_or(weights < 0, weights > 1).any():
        raise ValueError("weights must have probability between 0 and 1")

    weights = numpy.array(weights)
    samples = numpy.array(samples)

    state = numpy.random.get_state()

    numpy.random.seed(1)
    n = len(weights)
    choices = numpy.random.rand(n) < weights

    new_samples = samples[choices]

    numpy.random.set_state(state)

    return new_samples.copy()



import os
import pickle
import numpy
import inspect


class CacheException(Exception):
    """ Base exception to indicate cache errors """

    def calling_function(self):
        """ Get the name of the function calling this cache. """
        return inspect.getouterframes(inspect.currentframe())[3][3]

    def __str__(self):
        """ Return the cache message. """
        return self._msg


class CacheOK(CacheException):
    """ Exception to indicate the cache can be used. """
    def __init__(self, file_root):
        self._msg = "%s: reading from cache in %s" % (
                     self.calling_function(), file_root)


class CacheChanged(CacheException):
    """ Exception to indicate the cache has changed. """
    def __init__(self, file_root):
        self._msg = "%s: values have changed in cache %s, recomputing" % (
                     self.calling_function(), file_root)


class CacheMissing(CacheException):
    """ Exception to indicate the cache does not exist. """
    def __init__(self, file_root):
        self._msg = "%s: No cache file %s" % (
                     self.calling_function(), file_root)


class Cache(object):
    """ Cacheing tool for saving recomputation.

    Parameters
    ----------
    file_root: str
        cached values are saved in file_root.pkl
    """
    def __init__(self, file_root):
        self.file_root = file_root
        dirname = os.path.dirname(self.file_root)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def check(self, *args):
        """ Check that the arguments haven't changed since the last call.

        Parameters
        ----------
        *args:
            All but the last argument are inputs to the cached function. The
            last is the actual value of the function.

        Returns
        -------
        If arguments unchanged:
            return the cached answer
        else:
            indicate recomputation required by throwing a
            :class:`CacheException`.
        """
        data = self.load()

        if len(data)-1 != len(args):
            raise ValueError("Wrong number of arguments passed to Cache.check")

        try:
            for x, x_check in zip(data, args):
                if isinstance(x, list):
                    if len(x) != len(x_check):
                        raise CacheException
                    for x_i, x_check_i in zip(x, x_check):
                        if x_i.shape != x_check_i.shape:
                            raise CacheException
                        elif not numpy.allclose(x_i, x_check_i,
                                                equal_nan=True):
                            raise CacheException
                elif x.shape != x_check.shape:
                    raise CacheException
                elif not numpy.allclose(x, x_check, equal_nan=True):
                    raise CacheException

        except CacheException:
            raise CacheChanged(self.file_root)

        print(CacheOK(self.file_root))
        return data[-1]

    def load(self):
        """ Load cache from file using pickle. """
        try:
            with open(self.file_root + '.pkl', "rb") as f:
                return pickle.load(f)
        except IOError:
            raise CacheMissing(self.file_root)

    def save(self, *args):
        """ Save cache to file using pickle.

        Parameters
        ----------
        *args:
            All but the last argument are inputs to the cached function. The
            last is the actual value of the function.
        """
        with open(self.file_root + '.pkl', "wb") as f:
            pickle.dump(args, f, protocol=pickle.HIGHEST_PROTOCOL)



import numpy
import os
import pytest
from shutil import rmtree
from numpy.testing import assert_allclose, assert_almost_equal
from fgivenx.dkl import DKL, compute_dkl


def gaussian_dkl(mu, sig, mu_, sig_):
    return numpy.log(sig_/sig) + (sig**2+(mu-mu_)**2)/2/sig_**2 - 0.5


def test_DKL():
    numpy.random.seed(0)
    mu_, sig_ = 0, 1
    mu, sig = 0.2, 0.1
    size = 1000
    samples_ = numpy.random.normal(mu_, sig_, size)
    samples = numpy.random.normal(mu, sig, size)

    dkl = DKL((samples, samples_))
    dkl_true = gaussian_dkl(mu, sig, mu_, sig_)
    assert_almost_equal(dkl/dkl_true, 1, 1)


def test_compute_dkl():

    with pytest.raises(TypeError):
        compute_dkl(None, None, wrong_argument=None)

    cache = '.test_cache/test'
    numpy.random.seed(0)

    nx = 100
    x = numpy.linspace(-1, 1, nx)

    nsamp = 2000
    a, b, e, f = 0.1, 0.1, 0.1, 0.1
    m = numpy.random.normal(a, b, nsamp)
    c = numpy.random.normal(e, f, nsamp)
    fsamps = (numpy.outer(x, m) + c)

    a_, b_, e_, f_ = 0, 1, 0, 1
    m_ = numpy.random.normal(a_, b_, nsamp)
    c_ = numpy.random.normal(e_, f_, nsamp)
    prior_fsamps = (numpy.outer(x, m_) + c_)

    assert(not os.path.isfile(cache + '_dkl.pkl'))
    dkl = compute_dkl(fsamps, prior_fsamps, cache=cache)
    assert(os.path.isfile(cache + '_dkl.pkl'))

    dkl_ = [gaussian_dkl(a*xi+e,
                         numpy.sqrt(b**2*xi**2+f**2),
                         a_*xi+e_,
                         numpy.sqrt(b_**2*xi**2+f_**2)
                         ) for xi in x]
    assert_allclose(dkl, dkl_, atol=1e-1)

    dkl = compute_dkl(fsamps, prior_fsamps, cache=cache)
    assert_allclose(dkl, dkl_, atol=1e-1)

    rmtree('.test_cache')



"""

The main driving routines for this package are:

* :func:`plot_contours <fgivenx.drivers.plot_contours>`
* :func:`plot_lines <fgivenx.drivers.plot_lines>`
* :func:`plot_dkl <fgivenx.drivers.plot_dkl>`
* :func:`samples_from_getdist_chains
  <fgivenx.samples.samples_from_getdist_chains>`

Example import and usage:

>>> import numpy
>>> from fgivenx import plot_contours, plot_lines, \
...                     plot_dkl, samples_from_getdist_chains
>>>
>>> file_root = '/my/getdist/file/root'
>>> params = ['m', 'c']
>>> samples = samples_from_getdist_chains(params, file_root)
>>> x = numpy.linspace(-1, 1, 100)
>>>
>>> def f(x, theta):
>>>     m, c = params
>>>     y = m * x + c
>>>     return y
>>>
>>> plot_contours(f, x, samples)
"""

from fgivenx.drivers import plot_contours, plot_lines, plot_dkl
from fgivenx.samples import samples_from_getdist_chains



r"""This module provides utilities for computing the grid for contours of a
    function reconstruction plot.

    Required ingredients:
     * sampled posterior probability distribution :math:`P(\theta)`
     * independent variable :math:`x`
     * dependent variable :math:`y`
     * functional form :math:`y = f(x;\theta)` parameterised by :math:`\theta`

    Assuming that you have obtained samples of :math:`\theta` from an MCMC
    process, we aim to compute the density:

    .. math::

        P(y|x) &= \int P(y=f(x;\theta)|x,\theta) P(\theta) d\theta \\
               &= \int \delta(y-f(x;\theta)) P(\theta) d\theta

    which gives our degree of knowledge for each :math:`y=f(x;\theta)` value
    given an :math:`x` value.

    In fact, for a more representative plot, we are not actually
    interested in the value of the probability density above, but in fact
    require the "iso-probablity posterior mass"

    .. math::

        \mathrm{pmf}(y|x) = \int_{P(y'|x) < P(y|x)} P(y'|x) dy'

    We thus need to compute this function on a rectangular grid of :math:`x`
    and :math:`y`.
"""
import numpy
import fgivenx.samples
import fgivenx.mass
import fgivenx.dkl
import fgivenx.plot
import matplotlib.pyplot as plt
from fgivenx._utils import _check_args, _normalise_weights,\
                           _equally_weight_samples


def plot_contours(f, x, samples, ax=None, **kwargs):
    r"""
    Plot the probability mass function given `x` at a range of :math:`y` values
    for :math:`y = f(x|\theta)`

    :math:`P(y|x) = \int P(y=f(x;\theta)|x,\theta) P(\theta) d\theta`

    :math:`\mathrm{pmf}(y|x) = \int_{P(y'|x) < P(y|x)} P(y'|x) dy'`

    Additionally, if a list of log-evidences are passed, along with list of
    functions, and list of samples, this function plots the probability mass
    function for all models marginalised according to the evidences.

    Parameters
    ----------
    f: function
        function :math:`f(x;\theta)` (or list of functions for each model) with
        dependent variable :math:`x`, parameterised by :math:`\theta`.

    x: 1D array-like
        `x` values to evaluate :math:`f(x;\theta)` at.

    samples: 2D array-like
        :math:`\theta` samples (or list of :math:`\theta` samples) to evaluate
        :math:`f(x;\theta)` at.
        `shape = (nsamples, npars)`

    ax: axes object, optional
        :class:`matplotlib.axes._subplots.AxesSubplot` to plot the contours
        onto. If unsupplied, then :func:`matplotlib.pyplot.gca()` is used to
        get the last axis used, or create a new one.

    logZ: 1D array-like, optional
        log-evidences of each model if multiple models are passed.
        Should be same length as the list `f`, and need not be normalised.
        Default: `numpy.ones_like(f)`

    weights: 1D array-like, optional
        sample weights (or list of weights), if desired. Should have length
        same as `samples.shape[0]`.
        Default: `numpy.ones(samples.shape[0])`

    ny: int, optional
        Resolution of `y` axis.
        Default: `100`

    y: array-like, optional
        Explicit descriptor of `y` values to evaluate.
        Default: `numpy.linspace(min(f), max(f), ny)`

    ntrim: int, optional
        Approximate number of samples to trim down to, if desired. Useful if
        the posterior is dramatically oversampled.
        Default: None

    cache: str, optional
        File root for saving previous calculations for re-use

    parallel, tqdm_args:
        see docstring for :func:`fgivenx.parallel.parallel_apply`

    kwargs: further keyword arguments
        Any further keyword arguments are plotting keywords that are passed to
        :func:`fgivenx.plot.plot`.

    Returns
    -------
    cbar: color bar
        :class:`matplotlib.contour.QuadContourSet`
        Colors to create a global colour bar
    """
    logZ = kwargs.pop('logZ', None)
    weights = kwargs.pop('weights', None)
    ntrim = kwargs.pop('ntrim', None)
    ny = kwargs.pop('ny', 100)
    y = kwargs.pop('y', None)
    cache = kwargs.pop('cache', '')
    parallel = kwargs.pop('parallel', False)
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})

    y, pmf = compute_pmf(f, x, samples, weights=weights, logZ=logZ,
                         ntrim=ntrim, ny=ny, y=y,
                         parallel=parallel, cache=cache,
                         tqdm_kwargs=tqdm_kwargs)
    cbar = fgivenx.plot.plot(x, y, pmf, ax, **kwargs)
    return cbar


def plot_lines(f, x, samples, ax=None, **kwargs):
    r"""
    Plot a representative set of functions to sample

    Additionally, if a list of log-evidences are passed, along with list of
    functions, and list of samples, this function plots the probability mass
    function for all models marginalised according to the evidences.

    Parameters
    ----------
    f: function
        function :math:`f(x;\theta)` (or list of functions for each model) with
        dependent variable :math:`x`, parameterised by :math:`\theta`.

    x: 1D array-like
        `x` values to evaluate :math:`f(x;\theta)` at.

    samples: 2D array-like
        :math:`\theta` samples (or list of :math:`\theta` samples) to evaluate
        :math:`f(x;\theta)` at.
        `shape = (nsamples, npars)`

    ax: axes object, optional
        :class:`matplotlib.axes._subplots.AxesSubplot` to plot the contours
        onto. If unsupplied, then :func:`matplotlib.pyplot.gca()` is used to
        get the last axis used, or create a new one.

    logZ: 1D array-like, optional
        log-evidences of each model if multiple models are passed.
        Should be same length as the list `f`, and need not be normalised.
        Default: `numpy.ones_like(f)`

    weights: 1D array-like, optional
        sample weights (or list of weights), if desired. Should have length
        same as `samples.shape[0]`.
        Default: `numpy.ones(samples.shape[0])`

    ntrim: int, optional
        Approximate number of samples to trim down to, if desired. Useful if
        the posterior is dramatically oversampled.
        Default: None

    cache: str, optional
        File root for saving previous calculations for re-use

    parallel, tqdm_args:
        see docstring for :func:`fgivenx.parallel.parallel_apply`

    kwargs: further keyword arguments
        Any further keyword arguments are plotting keywords that are passed to
        :func:`fgivenx.plot.plot_lines`.
    """
    logZ = kwargs.pop('logZ', None)
    weights = kwargs.pop('weights', None)
    ntrim = kwargs.pop('ntrim', None)
    cache = kwargs.pop('cache', '')
    parallel = kwargs.pop('parallel', False)
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})

    fsamps = compute_samples(f, x, samples, logZ=logZ,
                             weights=weights, ntrim=ntrim,
                             parallel=parallel, cache=cache,
                             tqdm_kwargs=tqdm_kwargs)
    fgivenx.plot.plot_lines(x, fsamps, ax, **kwargs)


def plot_dkl(f, x, samples, prior_samples, ax=None, **kwargs):
    r"""
    Plot the Kullback-Leibler divergence at each value of :math:`x` for the
    prior and posterior defined by `prior_samples` and `samples`.

    Let the posterior be:

    :math:`P(y|x) = \int P(y=f(x;\theta)|x,\theta)P(\theta) d\theta`

    and the prior be:

    :math:`Q(y|x) = \int P(y=f(x;\theta)|x,\theta)Q(\theta) d\theta`

    then the Kullback-Leibler divergence at each x is defined by

    :math:`D_\mathrm{KL}(x)=\int P(y|x)\ln\left[\frac{Q(y|x)}{P(y|x)}\right]dy`

    Additionally, if a list of log-evidences are passed, along with list of
    functions, and list of samples, this function plots the Kullback-Leibler
    divergence for all models marginalised according to the evidences.

    Parameters
    ----------
    f: function
        function :math:`f(x;\theta)` (or list of functions for each model) with
        dependent variable :math:`x`, parameterised by :math:`\theta`.

    x: 1D array-like
        `x` values to evaluate :math:`f(x;\theta)` at.

    samples, prior_samples: 2D array-like
        :math:`\theta` samples (or list of :math:`\theta` samples) from
        posterior and prior to evaluate :math:`f(x;\theta)` at.
        `shape = (nsamples, npars)`

    ax: axes object, optional
        :class:`matplotlib.axes._subplots.AxesSubplot` to plot the contours
        onto. If unsupplied, then :func:`matplotlib.pyplot.gca()` is used to
        get the last axis used, or create a new one.

    logZ: 1D array-like, optional
        log-evidences of each model if multiple models are passed.
        Should be same length as the list `f`, and need not be normalised.
        Default: `numpy.ones_like(f)`

    weights, prior_weights: 1D array-like, optional
        sample weights (or list of weights), if desired. Should have length
        same as `samples.shape[0]`.
        Default: `numpy.ones(samples.shape[0])`

    ntrim: int, optional
        Approximate number of samples to trim down to, if desired. Useful if
        the posterior is dramatically oversampled.
        Default: None

    cache, prior_cache: str, optional
        File roots for saving previous calculations for re-use

    parallel, tqdm_args:
        see docstring for :func:`fgivenx.parallel.parallel_apply`

    kwargs: further keyword arguments
        Any further keyword arguments are plotting keywords that are passed to
        :func:`fgivenx.plot.plot`.
    """

    logZ = kwargs.pop('logZ', None)
    weights = kwargs.pop('weights', None)
    prior_weights = kwargs.pop('prior_weights', None)
    ntrim = kwargs.pop('ntrim', None)
    cache = kwargs.pop('cache', '')
    prior_cache = kwargs.pop('prior_cache', '')
    parallel = kwargs.pop('parallel', False)
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})

    dkls = compute_dkl(f, x, samples, prior_samples,
                       logZ=logZ, parallel=parallel,
                       cache=cache, prior_cache=prior_cache,
                       tqdm_kwargs=tqdm_kwargs,
                       ntrim=ntrim, weights=weights,
                       prior_weights=prior_weights)

    if ax is None:
        ax = plt.gca()
    ax.plot(x, dkls, **kwargs)


def compute_samples(f, x, samples, **kwargs):
    r"""
    Apply the function(s) :math:`f(x;\theta)` to the arrays defined in `x` and
    `samples`.  Has options for weighting, trimming, cacheing & parallelising.

    Additionally, if a list of log-evidences are passed, along with list of
    functions, samples and optional weights it marginalises over the models
    according to the evidences.

    Parameters
    ----------
    f: function
        function :math:`f(x;\theta)` (or list of functions for each model) with
        dependent variable :math:`x`, parameterised by :math:`\theta`.

    x: 1D array-like
        `x` values to evaluate :math:`f(x;\theta)` at.

    samples: 2D array-like
        :math:`\theta` samples (or list of :math:`\theta` samples) to evaluate
        :math:`f(x;\theta)` at.
        `shape = (nsamples, npars)`

    logZ: 1D array-like, optional
        log-evidences of each model if multiple models are passed.
        Should be same length as the list `f`, and need not be normalised.
        Default: `numpy.ones_like(f)`

    weights: 1D array-like, optional
        sample weights (or list of weights), if desired. Should have length
        same as `samples.shape[0]`.
        Default: `numpy.ones(samples.shape[0])`

    ntrim: int, optional
        Approximate number of samples to trim down to, if desired. Useful if
        the posterior is dramatically oversampled.
        Default: None

    cache: str, optional
        File root for saving previous calculations for re-use.
        Default: None

    parallel, tqdm_args:
        see docstring for :func:`fgivenx.parallel.parallel_apply`

    Returns
    -------
    2D numpy.array
        Evaluate the function `f` at each x value and each theta.
        Equivalent to `[[f(x_i,theta) for theta in samples] for x_i in x]`
    """
    logZ = kwargs.pop('logZ', None)
    weights = kwargs.pop('weights', None)
    ntrim = kwargs.pop('ntrim', None)
    cache = kwargs.pop('cache', '')
    parallel = kwargs.pop('parallel', False)
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})
    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    logZ, f, x, samples, weights = _check_args(logZ, f, x, samples, weights)

    logZ, weights = _normalise_weights(logZ, weights, ntrim)

    for i, (s, w) in enumerate(zip(samples, weights)):
        samples[i] = _equally_weight_samples(s, w)

    return fgivenx.samples.compute_samples(f, x, samples,
                                           parallel=parallel, cache=cache,
                                           tqdm_kwargs=tqdm_kwargs)


def compute_pmf(f, x, samples, **kwargs):
    r"""
    Compute the probability mass function given `x` at a range of `x` values
    for :math:`y = f(x|\theta)`

    :math:`P(y|x) = \int P(y=f(x;\theta)|x,\theta) P(\theta) d\theta`

    :math:`\mathrm{pmf}(y|x) = \int_{P(y'|x) < P(y|x)} P(y'|x) dy'`

    Additionally, if a list of log-evidences are passed, along with list of
    functions, samples and optional weights it marginalises over the models
    according to the evidences.

    Parameters
    ----------
    f: function
        function :math:`f(x;\theta)` (or list of functions for each model) with
        dependent variable :math:`x`, parameterised by :math:`\theta`.

    x: 1D array-like
        `x` values to evaluate :math:`f(x;\theta)` at.

    samples: 2D array-like
        :math:`\theta` samples (or list of :math:`\theta` samples) to evaluate
        :math:`f(x;\theta)` at.
        `shape = (nsamples, npars)`

    logZ: 1D array-like, optional
        log-evidences of each model if multiple models are passed.
        Should be same length as the list `f`, and need not be normalised.
        Default: `numpy.ones_like(f)`

    weights: 1D array-like, optional
        sample weights (or list of weights), if desired. Should have length
        same as `samples.shape[0]`.
        Default: `numpy.ones(samples.shape[0])`

    ny: int, optional
        Resolution of y axis.
        Default: `100`

    y: array-like, optional
        Explicit descriptor of `y` values to evaluate.
        Default: `numpy.linspace(min(f), max(f), ny)`

    ntrim: int, optional
        Approximate number of samples to trim down to, if desired. Useful if
        the posterior is dramatically oversampled.
        Default: None

    cache: str, optional
        File root for saving previous calculations for re-use

    parallel, tqdm_args:
        see docstring for :func:`fgivenx.parallel.parallel_apply`


    Returns
    -------
    1D numpy.array:
        `y` values pmf is computed at `shape=(len(y))` or `ny`
    2D numpy.array:
        pmf values at each `x` and `y`  `shape=(len(x),len(y))`
    """

    logZ = kwargs.pop('logZ', None)
    weights = kwargs.pop('weights', None)
    ny = kwargs.pop('ny', 100)
    y = kwargs.pop('y', None)
    ntrim = kwargs.pop('ntrim', 100000)
    parallel = kwargs.pop('parallel', False)
    cache = kwargs.pop('cache', '')
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})
    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    # y
    if y is not None:
        y = numpy.array(y, dtype='double')
        if len(y.shape) != 1:
            raise ValueError("y should be a 1D array")

    fsamps = compute_samples(f, x, samples, logZ=logZ,
                             weights=weights, ntrim=ntrim,
                             parallel=parallel, cache=cache,
                             tqdm_kwargs=tqdm_kwargs)

    if y is None:
        ymin = fsamps[~numpy.isnan(fsamps)].min(axis=None)
        ymax = fsamps[~numpy.isnan(fsamps)].max(axis=None)
        y = numpy.linspace(ymin, ymax, ny)

    return y, fgivenx.mass.compute_pmf(fsamps, y, parallel=parallel,
                                       cache=cache, tqdm_kwargs=tqdm_kwargs)


def compute_dkl(f, x, samples, prior_samples, **kwargs):
    r"""
    Compute the Kullback-Leibler divergence at each value of `x` for the prior
    and posterior defined by `prior_samples` and `samples`.

    Parameters
    ----------
    f: function
        function :math:`f(x;\theta)` (or list of functions for each model) with
        dependent variable :math:`x`, parameterised by :math:`\theta`.

    x: 1D array-like
        `x` values to evaluate :math:`f(x;\theta)` at.

    samples, prior_samples: 2D array-like
        :math:`\theta` samples (or list of :math:`\theta` samples) from
        posterior and prior to evaluate :math:`f(x;\theta)` at.
        `shape = (nsamples, npars)`

    logZ: 1D array-like, optional
        log-evidences of each model if multiple models are passed.
        Should be same length as the list `f`, and need not be normalised.
        Default: `numpy.ones_like(f)`

    weights, prior_weights: 1D array-like, optional
        sample weights (or list of weights), if desired. Should have length
        same as `samples.shape[0]`.
        Default: `numpy.ones(samples.shape[0])`

    ntrim: int, optional
        Approximate number of samples to trim down to, if desired. Useful if
        the posterior is dramatically oversampled.
        Default: None

    cache, prior_cache: str, optional
        File roots for saving previous calculations for re-use

    parallel, tqdm_args:
        see docstring for :func:`fgivenx.parallel.parallel_apply`

    kwargs: further keyword arguments
        Any further keyword arguments are plotting keywords that are passed to
        :func:`fgivenx.plot.plot`.

    Returns
    -------
    1D numpy array:
        dkl values at each value of `x`.
    """

    logZ = kwargs.pop('logZ', None)
    weights = kwargs.pop('weights', None)
    prior_weights = kwargs.pop('prior_weights', None)
    ntrim = kwargs.pop('ntrim', None)
    cache = kwargs.pop('cache', '')
    prior_cache = kwargs.pop('prior_cache', '')
    parallel = kwargs.pop('parallel', False)
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})

    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    if logZ is None:
        logZ = [0]
        f = [f]
        samples = [samples]
        prior_samples = [prior_samples]
        weights = [weights]
        prior_weights = [prior_weights]
        cache = [cache]
        prior_cache = [prior_cache]

    DKLs = []

    for fi, c, pc, s, w, ps, pw in zip(f, cache, prior_cache, samples, weights,
                                       prior_samples, prior_weights):

        fsamps = compute_samples(fi, x, s, weights=w, ntrim=ntrim,
                                 parallel=parallel, cache=c,
                                 tqdm_kwargs=tqdm_kwargs)

        fsamps_prior = compute_samples(fi, x, ps, weights=pw, ntrim=ntrim,
                                       parallel=parallel, cache=pc,
                                       tqdm_kwargs=tqdm_kwargs)

        dkls = fgivenx.dkl.compute_dkl(fsamps, fsamps_prior,
                                       parallel=parallel, cache=c,
                                       tqdm_kwargs=tqdm_kwargs)
        DKLs.append(dkls)

    logZ = numpy.array(logZ)
    DKLs = numpy.array(DKLs)

    Zs = numpy.exp(logZ-logZ.max())
    Zs /= Zs.sum()
    return numpy.sum(Zs * DKLs.transpose(), axis=1)



import numpy
import pytest
from fgivenx.plot import plot, plot_lines
import matplotlib.pyplot as plt
import matplotlib


def gen_plot_data():
    numpy.random.seed(0)
    nx = 100
    ny = 101
    x = numpy.linspace(0, 1, nx)
    y = numpy.linspace(0, 1, ny)
    X, Y = numpy.meshgrid(x, y)
    z = numpy.exp(-((X-0.5)**2+(Y-0.5)**2)/0.01)
    return x, y, z


def test_plot():
    x, y, z = gen_plot_data()

    fig, ax = plt.subplots()
    cbar = plot(x, y, z, ax)
    assert type(cbar) is matplotlib.contour.QuadContourSet


def test_plot_transparent():
    x, y, z = gen_plot_data()

    fig, ax = plt.subplots()
    cbar = plot(x, y, z, ax, alpha=0.7)
    assert type(cbar) is matplotlib.contour.QuadContourSet

def test_plot_fineness():
    x, y, z = gen_plot_data()

    fig, ax = plt.subplots()
    cbar = plot(x, y, z, ax, fineness=0.1, contour_line_levels=[1,2,3,4])
    assert type(cbar) is matplotlib.contour.QuadContourSet
    assert len(cbar.levels) == 41


def test_plot_wrong_argument():
    x, y, z = gen_plot_data()
    with pytest.raises(TypeError):
        plot(x, y, z, wrong_argument=None)


def test_plot_no_ax():
    plt.subplots()
    x, y, z = gen_plot_data()
    plot(x, y, z)


def test_plot_smooth():
    plt.subplots()
    x, y, z = gen_plot_data()
    plot(x, y, z, smooth=1)


def test_plot_rasterize():
    plt.subplots()
    x, y, z = gen_plot_data()
    plot(x, y, z, rasterize_contours=True)


def test_plot_nolines():
    plt.subplots()
    x, y, z = gen_plot_data()
    plot(x, y, z, lines=False)


def gen_line_data():
    numpy.random.seed(0)
    nx = 100
    nsamps = 150
    x = numpy.linspace(0, 1, nx)
    m = numpy.random.normal(size=nsamps)
    c = numpy.random.normal(size=nsamps)
    fsamps = numpy.outer(x, m) + c
    return x, fsamps


def test_plot_lines():
    x, fsamps = gen_line_data()
    fig, ax = plt.subplots()
    plot_lines(x, fsamps, ax)


def test_plot_lines_no_ax():
    x, fsamps = gen_line_data()
    plt.subplots()
    plot_lines(x, fsamps)


def test_plot_lines_downsample():
    x, fsamps = gen_line_data()
    plt.subplots()
    plot_lines(x, fsamps, downsample=200)



""" Utilities for computing the probability mass function. """
import scipy.stats
import scipy.interpolate
import numpy
from fgivenx.parallel import parallel_apply
from fgivenx.io import CacheException, Cache


def PMF(samples, y):
    """ Compute the probability mass function.

        The set of samples defines a probability density P(y),
        which is computed using a kernel density estimator.

        From :math:`P(y)` we define:

        :math:`\mathrm{pmf}(p) = \int_{P(y)<p} P(y) dy`

        This is the cumulative distribution function expressed as a
        function of the probability

        We aim to compute :math:`M(y)`, which indicates the amount of
        probability contained outside the iso-probability contour
        passing through :math:`y`::


             ^ P(y)                   ...
             |                     | .   .
             |                     |       .
            p|- - - - - - - - - - .+- - - - . - - - - - - - - - - -
             |                   .#|        #.
             |                  .##|        ##.
             |                  .##|        ##.
             |                 .###|        ###.     M(p)
             |                 .###|        ###.     is the
             |                 .###|        ###.     shaded area
             |                .####|        ####.
             |                .####|        ####.
             |              ..#####|        #####..
             |          ....#######|        #######....
             |         .###########|        ###########.
             +---------------------+-------------------------------> y
                                  t

             ^ M(p)                        ^ M(y)
             |                             |
            1|                +++         1|         +
             |               +             |        + +
             |       ++++++++              |       +   +
             |     ++                      |     ++     ++
             |   ++                        |   ++         ++
             |+++                          |+++             +++
             +---------------------> p     +---------------------> y
            0

        Parameters
        ----------
        samples: array-like
            Array of samples from a probability density P(y).

        y: array-like (optional)
            Array to evaluate the PDF at

        Returns
        -------
        1D numpy.array:
            PMF evaluated at each y value

    """
    # Remove any nans from the samples
    samples = numpy.array(samples)
    samples = samples[~numpy.isnan(samples)]
    try:
        # Compute the kernel density estimate
        kernel = scipy.stats.gaussian_kde(samples)

        # Add two more samples definitely outside the range and sort them
        mn = min(samples) - 10*numpy.sqrt(kernel.covariance[0, 0])
        mx = max(samples) + 10*numpy.sqrt(kernel.covariance[0, 0])
        y_ = numpy.linspace(mn, mx, len(y)*10)

        # Compute the probabilities at each of the extended samples
        ps_ = kernel(y_)

        # Compute the masses
        ms = []
        for yi in y:
            # compute the probability at this y value
            p = kernel(yi)
            if p <= max(ps_)*1e-5:
                m = 0.
            else:
                # Find out which samples have greater probability than P(y)
                bools = ps_ > p

                # Compute indices where to start and stop the integration
                stops = numpy.where(numpy.logical_and(~bools[:-1], bools[1:]))
                starts = numpy.where(numpy.logical_and(bools[:-1], ~bools[1:]))

                # Compute locations
                starts = [scipy.optimize.brentq(lambda u: kernel(u)-p,
                                                y_[i], y_[i+1])
                          for i in starts[0]]
                starts = [-numpy.inf] + starts
                stops = [scipy.optimize.brentq(lambda u: kernel(u)-p,
                                               y_[i], y_[i+1])
                         for i in stops[0]]
                stops = stops + [numpy.inf]

                # Sum up the masses
                m = sum(kernel.integrate_box_1d(a, b)
                        for a, b in zip(starts, stops))
            ms.append(m)
        return numpy.array(ms)

    except numpy.linalg.LinAlgError:
        return numpy.zeros_like(y)


def compute_pmf(fsamps, y, **kwargs):
    """ Compute the pmf defined by fsamps at each x for each y.

    Parameters
    ----------
    fsamps: 2D array-like
        array of function samples, as returned by
        :func:`fgivenx.compute_samples`

    y: 1D array-like
        y values to evaluate the PMF at

    parallel, tqdm_kwargs: optional
        see docstring for :func:`fgivenx.parallel.parallel_apply`.

    Returns
    -------
    2D numpy.array
        probability mass function at each x for each y
        `shape=(len(fsamps),len(y)`
    """
    parallel = kwargs.pop('parallel', False)
    cache = kwargs.pop('cache', '')
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})
    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    if cache:
        cache = Cache(cache + '_masses')
        try:
            return cache.check(fsamps, y)
        except CacheException as e:
            print(e)

    masses = parallel_apply(PMF, fsamps, postcurry=(y,), parallel=parallel,
                            tqdm_kwargs=tqdm_kwargs)
    masses = numpy.array(masses).transpose().copy()

    if cache:
        cache.save(fsamps, y, masses)

    return masses



import numpy
from scipy.stats import gaussian_kde
from fgivenx.io import CacheException, Cache
from fgivenx.parallel import parallel_apply


def DKL(arrays):
    """
    Compute the Kullback-Leibler divergence from one distribution Q to another
    P, where Q and P are represented by a set of samples.

    Parameters
    ----------
    arrays: tuple(1D numpy.array,1D numpy.array)
        samples defining distributions P & Q respectively

    Returns
    -------
    float:
        Kullback Leibler divergence.
    """
    samples, prior_samples = arrays
    samples = samples[~numpy.isnan(samples)]
    prior_samples = prior_samples[~numpy.isnan(prior_samples)]
    return (
            gaussian_kde(samples).logpdf(samples)
            - gaussian_kde(prior_samples).logpdf(samples)
            ).mean()


def compute_dkl(fsamps, prior_fsamps, **kwargs):
    """
    Compute the Kullback Leibler divergence for function samples for posterior
    and prior pre-calculated at a range of x values.

    Parameters
    ----------
    fsamps: 2D numpy.array
        Posterior function samples, as computed by
        :func:`fgivenx.compute_samples`

    prior_fsamps: 2D numpy.array
        Prior function samples, as computed by :func:`fgivenx.compute_samples`

    parallel, tqdm_kwargs: optional
        see docstring for :func:`fgivenx.parallel.parallel_apply`.

    cache: str, optional
        File root for saving previous calculations for re-use.

    Returns
    -------
    1D numpy.array:
        Kullback-Leibler divergences at each value of x. `shape=(len(fsamps))`
    """

    parallel = kwargs.pop('parallel', False)
    cache = kwargs.pop('cache', '')
    tqdm_kwargs = kwargs.pop('tqdm_kwargs', {})
    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    if cache:
        cache = Cache(cache + '_dkl')
        try:
            return cache.check(fsamps, prior_fsamps)
        except CacheException as e:
            print(e)

    zip_fsamps = list(zip(fsamps, prior_fsamps))
    dkls = parallel_apply(DKL, zip_fsamps, parallel=parallel,
                          tqdm_kwargs=tqdm_kwargs)
    dkls = numpy.array(dkls)

    if cache:
        cache.save(fsamps, prior_fsamps, dkls)

    return dkls



#!/usr/bin/env python3

try:
    from setuptools import setup, Command
except ImportError:
    from distutils.core import setup, Command


def readme():
    with open('README.rst') as f:
        return f.read()


def get_version(short=False):
    with open('README.rst') as f:
        for line in f:
            if ':Version:' in line:
                ver = line.split(':')[2].strip()
                if short:
                    subver = ver.split('.')
                    return '%s.%s' % tuple(subver[:2])
                else:
                    return ver


setup(name='fgivenx',
      version=get_version(),
      description='fgivenx: Functional Posterior Plotter',
      long_description=readme(),
      author='Will Handley',
      author_email='wh260@cam.ac.uk',
      url='https://github.com/fgivenx/fgivenx',
      packages=['fgivenx', 'fgivenx.test'],
      install_requires=['matplotlib', 'numpy', 'scipy'],
      setup_requires=['pytest-runner'],
      extras_require={
          'docs': ['sphinx', 'sphinx_rtd_theme', 'numpydoc'],
          'parallel': ['joblib'],
          'progress_bar': ['tqdm'],
          'getdist_chains': ['getdist']
          },
      tests_require=['pytest', 'pytest-mpl'],
      include_package_data=True,
      license='MIT',
      classifiers=[
                   'Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'Natural Language :: English',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Topic :: Scientific/Engineering',
                   'Topic :: Scientific/Engineering :: Astronomy',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Topic :: Scientific/Engineering :: Visualization',
                   'Topic :: Scientific/Engineering :: Information Analysis',
                   'Topic :: Scientific/Engineering :: Mathematics',
      ],
      )



import numpy
import pytest
from numpy.testing import assert_allclose
from fgivenx.parallel import parallel_apply


def f(x):
    return x**3


def test_parallel_apply():
    array = numpy.linspace(-1, 1, 100)

    with pytest.raises(TypeError):
        parallel_apply(f, array, wrong_argument=None)

    with pytest.raises(ValueError):
        parallel_apply(f, array, parallel='wrong_argument')

    for par in [False, True, -1, 2]:
        farray = parallel_apply(f, array, parallel=par)
        assert_allclose([f(x) for x in array], farray)
